#pragma once
#include <math.h>
#include <string>
#include <vector>
namespace Thorbjorn {
	namespace Math {



#define _PI 3.14159265358f


		inline float toRadians(float degrees)
		{
			return (float)(degrees * (_PI / 180.0f));
		}

		inline float toDegrees(float radians)
		{
			return (float)(radians * (180.0f / _PI));
		}

		inline int sign(float value)
		{
			return (value > 0) - (value < 0);
		}

		inline float sin(float angle)
		{
			return ::sin(angle);
		}

		inline float cos(float angle)
		{
			return ::cos(angle);
		}

		inline float tan(float angle)
		{
			return ::tan(angle);
		}

		inline float sqrt(float value)
		{
			return ::sqrt(value);
		}

		inline float rsqrt(float value)
		{
			return 1.0f / ::sqrt(value);
		}

		inline float asin(float value)
		{
			return ::asin(value);
		}

		inline float acos(float value)
		{
			return ::acos(value);
		}

		inline float atan(float value)
		{
			return ::atan(value);
		}

		inline float atan2(float y, float x)
		{
			return ::atan2(y, x);
		}

		inline float _min(float value, float minimum)
		{
			return (value < minimum) ? minimum : value;
		}

		inline float _max(float value, float maximum)
		{
			return (value > maximum) ? maximum : value;
		}

		inline float clamp(float value, float minimum, float maximum)
		{
			return (value > minimum) ? (value < maximum) ? value : maximum : minimum;
		}
		inline void Tokenize(std::vector<std::string>* tokens, std::string s, std::string delimiter)
		{
			// check that is still valid
			while (s.find(delimiter, 0) != std::string::npos) {
				// get pos of next delimiter
				size_t pos = s.find(delimiter, 0);
				// create a token
				std::string token = s.substr(0, pos);
				// erase the current token 
				s.erase(0, pos + 1);
				// add the token to  vector
				tokens->push_back(token);
			}
			// push the last token
			tokens->push_back(s);
		}
		struct  vec2;
		struct  vec4;
		struct  vec3;
		struct mat4;
		using uint = unsigned int;
		using int32 = int;
		

		struct Quaternion;
		struct  Quaternion
		{
			float x, y, z, w;

			Quaternion();
			Quaternion(const Quaternion& quaternion);
			Quaternion(float x, float y, float z, float w);
			Quaternion(const vec3& xyz, float w);
			Quaternion(const vec4& vec);
			Quaternion(float scalar);

			Quaternion& operator=(const Quaternion& quat);

			Quaternion& SetXYZ(const vec3& vec);
			const vec3 GetXYZ() const;

			Quaternion& SetElem(int32 idx, float value);
			float GetElem(int32 idx) const;
			vec3 GetAxis() const;
			vec3 ToEulerAngles() const;

			const Quaternion operator+(const Quaternion& Quaternion) const;
			const Quaternion operator-(const Quaternion& Quaternion) const;
			const Quaternion operator*(const Quaternion& Quaternion) const;
			const Quaternion operator*(float scalar) const;
			const Quaternion operator/(float scalar) const;
			float operator[](int32 idx) const;

			Quaternion& operator+=(const Quaternion& Quaternion)
			{
				*this = *this + Quaternion;
				return *this;
			}

			Quaternion& operator-=(const Quaternion& Quaternion)
			{
				*this = *this - Quaternion;
				return *this;
			}

			Quaternion& operator*=(const Quaternion& Quaternion)
			{
				*this = *this * Quaternion;
				return *this;
			}

			Quaternion& operator*=(float scalar)
			{
				*this = *this * scalar;
				return *this;
			}

			Quaternion& operator/=(float scalar)
			{
				*this = *this / scalar;
				return *this;
			}

			const Quaternion operator-() const;
			bool operator==(const Quaternion& quaternion) const;
			bool operator!=(const Quaternion& quaternion) const;
			static Quaternion Identity();
			static Quaternion FromEulerAngles(const vec3& angles);

			static vec3 Rotate(const Quaternion & quat, const vec3 & vec);

			static const Quaternion Rotation(const vec3& unitVec0, const vec3& unitVec1);
			static const Quaternion Rotation(float radians, const vec3& unitVec);

			static const Quaternion RotationX(float radians)
			{
				float angle = radians * 0.5f;
				return Quaternion(sin(angle), 0.0f, 0.0f, cos(angle));
			}

			static const Quaternion RotationY(float radians)
			{
				float angle = radians * 0.5f;
				return Quaternion(0.0f, sin(angle), 0.0f, cos(angle));
			}

			static const Quaternion RotationZ(float radians)
			{
				float angle = radians * 0.5f;
				return Quaternion(0.0f, 0.0f, sin(angle), cos(angle));
			}

			float Dot(const Quaternion& other) const;
			Quaternion Conjugate() const;

		};
		struct  vec4
		{
			union {
				struct {
					float x, y, z, w;
				};
				struct {
					float r, g, b, a;
				};
				struct {
					float left, top, right, bottom;
				};
			};


			vec4() = default;
			vec4(float scalar);
			vec4(float x, float y, float z, float w);
			vec4(const vec3& xyz, float w);



			vec4& Add(const vec4& other);
			vec4& Subtract(const vec4& other);
			vec4& Multiply(const vec4& other);
			vec4& Divide(const vec4& other);

			vec4 Multiply(const mat4& transform) const;

			friend vec4 operator+(vec4 left, const vec4& right);
			friend vec4 operator-(vec4 left, const vec4& right);
			friend vec4 operator*(vec4 left, const vec4& right);
			friend vec4 operator/(vec4 left, const vec4& right);

			bool operator==(const vec4& other);
			bool operator!=(const vec4& other);

			vec4& operator+=(const vec4& other);
			vec4& operator-=(const vec4& other);
			vec4& operator*=(const vec4& other);
			vec4& operator/=(const vec4& other);

			float Dot(const vec4& other);
			bool overlaps(const vec4& rect);
			friend std::ostream& operator<<(std::ostream& stream, const vec4& vector);
		};
		struct  mat4
		{
			union
			{
				// [row + col * 4]
				float elements[4 * 4];
				vec4 rows[4];
			};

			mat4();
			mat4(float diagonal);
			mat4(float* elements);
			mat4(const vec4& row0, const vec4& row1, const vec4& row2, const vec4& row3);

			static mat4 Identity();

			mat4& Multiply(const mat4& other);
			friend  mat4 operator*(mat4 left, const mat4& right);
			mat4& operator*=(const mat4& other);

			vec3 Multiply(const vec3& other) const;
			friend  vec3 operator*(const mat4& left, const vec3& right);

			vec4 Multiply(const vec4& other) const;
			friend  vec4 operator*(const mat4& left, const vec4& right);

			mat4& Invert();

			vec4 GetColumn(int index) const;
			void SetColumn(uint index, const vec4& column);
			vec3 GetPosition() const;
			void SetPosition(const vec3& position);

			static mat4 Orthographic(float left, float right, float bottom, float top, float _near, float _far);
			static mat4 Perspective(float fov, float aspectRatio, float _near, float _far);
			static mat4 LookAt(const vec3& camera, const vec3& object, const vec3& up);

			static mat4 Translate(const vec3& translation);
			static mat4 Rotate(float angle, const vec3& axis);
			static mat4 Rotate(const Quaternion& quat);
			static mat4 Scale(const vec3& scale);
			static mat4 Invert(const mat4& matrix);

			static mat4 Transpose(const mat4& matrix);

			//String ToString() const;
		};

		

		struct  vec3
		{
			float x, y, z;

			vec3();
			vec3(float scalar);
			vec3(float x, float y, float z);
			vec3(const vec2& other);
			vec3(float x, float y);
			vec3(const vec4& other);

			static vec3 Up();
			static vec3 Down();
			static vec3 Left();
			static vec3 Right();
			static vec3 Zero();

			static vec3 XAxis();
			static vec3 YAxis();
			static vec3 ZAxis();

			vec3& Add(const vec3& other);
			vec3& Subtract(const vec3& other);
			vec3& Multiply(const vec3& other);
			vec3& Divide(const vec3& other);

			vec3& Add(float other);
			vec3& Subtract(float other);
			vec3& Multiply(float other);
			vec3& Divide(float other);



			friend vec3 operator+(vec3 left, const vec3& right);
			friend vec3 operator-(vec3 left, const vec3& right);
			friend vec3 operator*(vec3 left, const vec3& right);
			friend vec3 operator/(vec3 left, const vec3& right);

			friend vec3 operator+(vec3 left, float right);
			friend vec3 operator-(vec3 left, float right);
			friend vec3 operator*(vec3 left, float right);
			friend vec3 operator/(vec3 left, float right);

			bool operator==(const vec3& other) const;
			bool operator!=(const vec3& other) const;

			vec3& operator+=(const vec3& other);
			vec3& operator-=(const vec3& other);
			vec3& operator*=(const vec3& other);
			vec3& operator/=(const vec3& other);

			vec3& operator+=(float other);
			vec3& operator-=(float other);
			vec3& operator*=(float other);
			vec3& operator/=(float other);

			bool operator<(const vec3& other) const;
			bool operator<=(const vec3& other) const;
			bool operator>(const vec3& other) const;
			bool operator>=(const vec3& other) const;

			friend vec3 operator-(const vec3& vector);

			vec3 Cross(const vec3& other) const;
			float Dot(const vec3& other) const;

			float Magnitude() const;
			vec3 Normalize() const;
			float Distance(const vec3& other) const;
			vec3 Multiply(const mat4& transform) const;

		};

		struct  vec2
		{
			union {
				struct {
					float x;
					float y;
				};
				struct {
					float width;
					float height;
				};
				struct {
					float min;
					float max;
				};
			};

			vec2();
			vec2(float scalar);
			vec2(float x, float y);


			vec2& Add(const vec2& other);
			vec2& Subtract(const vec2& other);
			vec2& Multiply(const vec2& other);
			vec2& Divide(const vec2& other);

			vec2& Add(float value);
			vec2& Subtract(float value);
			vec2& Multiply(float value);
			vec2& Divide(float value);

			friend vec2 operator+(vec2 left, const vec2& right);
			friend vec2 operator-(vec2 left, const vec2& right);
			friend vec2 operator*(vec2 left, const vec2& right);
			friend vec2 operator/(vec2 left, const vec2& right);

			friend vec2 operator+(vec2 left, float value);
			friend vec2 operator-(vec2 left, float value);
			friend vec2 operator*(vec2 left, float value);
			friend vec2 operator/(vec2 left, float value);

			bool operator==(const vec2& other) const;
			bool operator!=(const vec2& other) const;

			vec2& operator+=(const vec2& other);
			vec2& operator-=(const vec2& other);
			vec2& operator*=(const vec2& other);
			vec2& operator/=(const vec2& other);

			vec2& operator=(const vec3& other);

			vec2& operator+=(float value);
			vec2& operator-=(float value);
			vec2& operator*=(float value);
			vec2& operator/=(float value);

			bool operator<(const vec2& other) const;
			bool operator<=(const vec2& other) const;
			bool operator>(const vec2& other) const;
			bool operator>=(const vec2& other) const;

			float Magnitude() const;
			vec2 Normalise() const;
			float Distance(const vec2& other) const;
			float Dot(const vec2& other) const;



			//String toString();
			//friend std::ostream& operator<<(std::ostream& stream, const vec2& vector);
		};

		
		
		
		inline vec3 vec3Max(vec3 a, vec3 b)
		{
			float va[3] = { a.x,a.y,a.z };
			float vb[3] = { b.x,b.y,b.z };
			float result[3];
			for (int i = 0; i < 3; i++)
			{
				result[i] = va[i] < vb[i] ? va[i] : vb[i];
			}
			return vec3(result[0], result[1], result[2]);
		};

		inline float MaxElement(const vec3& vec)
		{
			float val[3] = { vec.x,vec.y,vec.z };
			float maxVal = val[0];
			for (int i = 0; i < 3; i++)
			{
				if (val[i] > maxVal)
					maxVal = val[i];
			}
			return maxVal;
		}
		inline float triangle_area(vec3 p1, vec3 p2, vec3 p3)
		{
			//area of the triangle with the heron fomula
			float a = sqrtf((p2.x - p1.x)*(p2.x - p1.x) + (p2.y - p1.y)*(p2.y - p1.y) + (p2.z - p1.z)*(p2.z - p1.z));
			float b = sqrtf((p3.x - p2.x)*(p3.x - p2.x) + (p3.y - p2.y)*(p3.y - p2.y) + (p3.z - p2.z)*(p3.z - p2.z));
			float c = sqrtf((p1.x - p3.x)*(p1.x - p3.x) + (p1.y - p3.y)*(p1.y - p3.y) + (p1.z - p3.z)*(p1.z - p3.z));
			float s = (a + b + c) / 2.0f;
			return (sqrtf(s*((s - a)*(s - b)*(s - c))));

		}
		// n = -plane normal s = sphere center d = normal
		inline bool rayplane(float nx, float ny, float nz, float xs, float ys, float zs, float xd, float yd, float zd, vec3 p1, vec3 p2, vec3 p3, float* dist, vec3* point)
		{
			float a = xd * nx + yd * ny + zd * nz;
			if (a == 0.0f)
				return false;
			float t = ((p1.x*nx + p1.y*ny + p1.z*nz - nx * xs - ny * ys - nz * zs) / a);
			if (t < 0.0f)
				return false;
			float x = xs + t * xd;
			float y = ys + t * yd;
			float z = zs + t * zd;
			vec3 cp(x, y, z);
			float ta = triangle_area(p1, p2, p3);
			float res = fabsf(triangle_area(p1, p2, p3) - triangle_area(p1, p2, cp) - triangle_area(p2, p3, cp) - triangle_area(p1, p3, cp));
			if (res < 0.000001f)
			{
				if (dist != NULL)
				{
					(*dist) = t;
					if (point != NULL)
					{
						point->x = x;
						point->y = y;
						point->z = z;
					}
				}
				return true;
			}
			return false;
		}

	}
}

